/*
 * org.nrg.xft.schema.XFTManager
 * XNAT http://www.xnat.org
 * Copyright (c) 2014, Washington University School of Medicine
 * All Rights Reserved
 *
 * Released under the Simplified BSD.
 *
 * Last modified 8/28/13 3:19 PM
 */
package org.nrg.xft.schema;

import java.io.File;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;

import com.google.common.base.Joiner;
import org.nrg.xft.XFT;
import org.nrg.xft.collections.XFTElementSorter;
import org.nrg.xft.db.DBConfig;
import org.nrg.xft.db.DBPool;
import org.nrg.xft.db.ViewManager;
import org.nrg.xft.exception.ElementNotFoundException;
import org.nrg.xft.exception.XFTInitException;
import org.nrg.xft.references.XFTReferenceI;
import org.nrg.xft.references.XFTReferenceManager;
import org.nrg.xft.references.XFTRelationSpecification;
import org.nrg.xft.references.XFTSuperiorReference;
import org.nrg.xft.schema.Wrappers.GenericWrapper.GenericWrapperElement;
import org.nrg.xft.schema.Wrappers.GenericWrapper.GenericWrapperFactory;
import org.nrg.xft.schema.Wrappers.GenericWrapper.GenericWrapperField;
import org.nrg.xft.schema.Wrappers.XMLWrapper.XMLWriter;
import org.nrg.xft.utils.FileUtils;
import org.nrg.xft.utils.NodeUtils;
import org.nrg.xft.utils.XMLUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

public class XFTManager {
    private static final Logger _log = LoggerFactory.getLogger(XFTManager.class);

    private static XFTManager MANAGER = null;
    private static XFTElement ELEMENT_TABLE = null;
    private static Hashtable DATA_MODELS = new Hashtable();
    private static Hashtable ROOT_LEVEL_ELEMENTS= new Hashtable();

    //private String packageName = "";
    private String sourceDir = "";

    /**
     * Gets singleton instance of the Manager
     * @return An instance of the XFT manager class
     * @throws XFTInitException
     */
    public static XFTManager GetInstance() throws XFTInitException
    {
        if (MANAGER == null)
        {
            throw new XFTInitException();
        }
        return MANAGER;
    }

    /**
     * Initializes the XFT manager instance if it hasn't been already.
     * @param schemaLocation    Path to the root folder containing the schema files.
     * @return The initialized XFT manager.
     * @throws ElementNotFoundException
     */
    public static XFTManager init(String schemaLocation) throws ElementNotFoundException
    {
        if (_log.isDebugEnabled()) {
            _log.debug("Initializing the XFT manager using the schema location: {}", schemaLocation);
        }

        MANAGER = new XFTManager(schemaLocation);

        if (_log.isDebugEnabled()) {
            _log.debug("Initialized the XFT manager. Found the following:");
            _log.debug(" * Source dir: {}", MANAGER.getSourceDir());
            _log.debug(" * {} add-in elements", MANAGER.getAddInElements().size());
            try {
                _log.debug(" * {} add-in elements", MANAGER.getOrderedElements().size());
            } catch (Exception e) {
                _log.error(" * An error occurred trying to retrieve the ordered elements", e);
            }
            try {
                _log.debug(" * {} all elements", MANAGER.getAllElements().size());
            } catch (Exception e) {
                _log.error(" * An error occurred trying to retrieve all elements", e);
            }
        }

        //XFT.LogCurrentTime("MANAGER INIT:2","ERROR");
        try {
            MANAGER.manageAddins();
        }catch (XFTInitException e) {
            _log.error("An initialization exception occurred trying to manage the add-ins", e);
        }catch (Exception e) {
            _log.error("A general exception occurred trying to manage the add-ins", e);
        }

        _log.debug("We seem to have made it through the constructor.");
        return MANAGER;
    }

    public static void clean()
    {
        _log.debug("Cleaning out the XFT manager data stores.");
        MANAGER = null;
        ELEMENT_TABLE = null;
        DATA_MODELS = new Hashtable();
    }

    /**
     * Returns the XFT element that stores all of the element names.
     * @return The XFT element with all of the element names.
     */
    public static XFTElement GetElementTable()
    {
        return ELEMENT_TABLE;
    }

    /**
     * Sets the XFTElement which stores all of the element names.
     * @param xe    The XFT element with all of the element names to set.
     */
    public static void SetElementTable(XFTElement xe)
    {
        ELEMENT_TABLE = xe;
    }

    /**
     * Re-initializes the XFT manager
     * @param sourceDirectory    The source directory from which to load the data schema.
     * @throws XFTInitException
     */
    @SuppressWarnings("unused")
    public static void Refresh(String sourceDirectory) throws XFTInitException,ElementNotFoundException
    {
        if (_log.isDebugEnabled()) {
            _log.debug("Refreshing the XFT manager from the source directory: {}", sourceDirectory);
        }
        MANAGER = null;
        init(sourceDirectory);
    }

    /**
     * Gets the XFT schema documents contained in the collection of data models.
     * @return A list of all of the XFT schema documents.
     * @see XFTSchema
     */
    public static ArrayList GetSchemas()
    {
        _log.debug("Getting the list of data-model schema.");
        ArrayList al = new ArrayList();
        Enumeration enumer = DATA_MODELS.keys();
        while(enumer.hasMoreElements())
        {
            al.add(((XFTDataModel)DATA_MODELS.get(enumer.nextElement())).getSchema());
        }
        if (_log.isDebugEnabled()) {
            _log.debug("Retrieved a total of {} data models.", al.size());
        }
        return al;
    }
    /**
     * Gets the XFTDataModels
     * @return hash of XFTDataModels
     * @see XFTDataModel
     */
    public static Hashtable GetDataModels() {
        return DATA_MODELS;
    }
    /**
     * Gets the Root Elements
     * @return hash of String properName,String complexType
     */
    public static Hashtable GetRootElementsHash() {
        return ROOT_LEVEL_ELEMENTS;
    }
    /**
     * Add a Root Element
     */
    public static void AddRootElement(String name, String elementName) {
        if (_log.isDebugEnabled()) {
            _log.debug("Adding {} with root element {}", name, elementName);
        }
        ROOT_LEVEL_ELEMENTS.put(name,elementName);
    }
    /**
     * Source directory where the InstanceSettings.xml document can be found.
     * @return The path to the source directory.
     */
    public String getSourceDir() {
        return sourceDir;
    }

    /**
     * Source directory where the InstanceSettings.xml document can be found.
     * @param sourceDir    The source directory to be set.
     */

    @SuppressWarnings("unused")
    public void setSourceDir(String sourceDir) {
        if (_log.isDebugEnabled()) {
            _log.debug("Setting the source directory to {}", sourceDir);
        }
        this.sourceDir = sourceDir;
    }

    /**
     * Access the InstanceSettings.xml document, and parses it to create a
     * collection of DB Connections in the DBPool and a collection of XFTDataModels (local).
     * @param source location where InstanceSettings.xml can be found.
     * @throws ElementNotFoundException
     */
    private XFTManager(String source) throws ElementNotFoundException
    {
        _log.debug("Java Version is: " + System.getProperty("java.version"));
        if (! source.endsWith(File.separator))
        {
            source = source + File.separator;
        }
        if (source.contains("WEB-INF")){
            sourceDir = source.substring(0,source.indexOf("WEB-INF"));
            System.out.println("SOURCE: " + sourceDir);
        }else{
            sourceDir = source;
        }
        File file = new File(source + "InstanceSettings.xml");
        if (_log.isDebugEnabled()) {
            _log.debug("Initializing the XFT manager from the instance settings stored in {}. This file {}.", file.getAbsolutePath(), file.exists() ? "exists" : "does not exist");
        }
        Document doc = XMLUtils.GetDOM(file);
        Element root = doc.getDocumentElement();

        ViewManager.PRE_LOAD_HISTORY = NodeUtils.GetBooleanAttributeValue(root,"Pre_Load_History",false);

        String admin_email = (NodeUtils.GetAttributeValue(root,"admin_email",""));
        if (!admin_email.equals(""))
        {
            if (_log.isDebugEnabled()) {
                _log.debug("Setting the admin email to {}", admin_email);
            }
            XFT.SetAdminEmail(admin_email);
        }
        String site_url = (NodeUtils.GetAttributeValue(root,"site_url",""));
        if (!site_url.equals(""))
        {
            if (_log.isDebugEnabled()) {
                _log.debug("Setting the site URL to {}", site_url);
            }
            XFT.SetSiteURL(site_url);
        }

        String archive_root_path = (NodeUtils.GetAttributeValue(root,"archive_root_path",""));
        if (!archive_root_path.equals(""))
        {
            if (_log.isDebugEnabled()) {
                _log.debug("Setting the archive root path to {}", archive_root_path);
            }
            XFT.SetArchiveRootPath(archive_root_path);
        }

        String prearchive_path = (NodeUtils.GetAttributeValue(root,"prearchive_path",""));
        if (!prearchive_path.equals(""))
        {
            if (_log.isDebugEnabled()) {
                _log.debug("Setting the prearchive path to {}", prearchive_path);
            }
            XFT.SetPrearchivePath(prearchive_path);
        }

        String cache_path = (NodeUtils.GetAttributeValue(root,"cache_path",""));
        if (!cache_path.equals(""))
        {
            if (_log.isDebugEnabled()) {
                _log.debug("Setting the cache path to {}", cache_path);
            }
            XFT.SetCachePath(cache_path);
        }

        String smtp_server = (NodeUtils.GetAttributeValue(root,"smtp_server",""));
        if (!smtp_server.equals(""))
        {
            if (_log.isDebugEnabled()) {
                _log.debug("Setting the SMTP server to {}", smtp_server);
            }
            XFT.SetAdminEmailHost(smtp_server);
        }

        String pipeline_path = (NodeUtils.GetAttributeValue(root,"pipeline_path",""));
        if (!pipeline_path.equals(""))
        {
            if (_log.isDebugEnabled()) {
                _log.debug("Setting the pipeline path to {}", pipeline_path);
            }
            XFT.SetPipelinePath(pipeline_path);
        }

        String ftp_path = (NodeUtils.GetAttributeValue(root,"ftp_path",""));
        if (!ftp_path.equals(""))
        {
            if (_log.isDebugEnabled()) {
                _log.debug("Setting the ftp path to {}", ftp_path);
            }
            XFT.setFtpPath(ftp_path);
        }

        String build_path = (NodeUtils.GetAttributeValue(root,"build_path",""));
        if (!build_path.equals(""))
        {
            if (_log.isDebugEnabled()) {
                _log.debug("Setting the build path to {}", build_path);
            }
            XFT.setBuildPath(build_path);
        }

        String require_login = (NodeUtils.GetAttributeValue(root,"require_login",""));
        if (!require_login.equals(""))
        {
            if (_log.isDebugEnabled()) {
                _log.debug("Setting require login to {}", require_login);
            }
            XFT.SetRequireLogin(require_login);
        }

        String user_registration = (NodeUtils.GetAttributeValue(root,"user_registration",""));
        if (!user_registration.equals(""))
        {
            if (_log.isDebugEnabled()) {
                _log.debug("Setting user registration to {}", user_registration);
            }
            XFT.SetUserRegistration(user_registration);
        }

        if (root.hasChildNodes())
        {
            for (int i=0;i<root.getChildNodes().getLength();i++)
            {
                Node child1 = root.getChildNodes().item(i);
                if (child1.getNodeName().equalsIgnoreCase("Databases"))
               {
                   if (child1.hasChildNodes())
                   {
                       for (int j=0;j<child1.getChildNodes().getLength();j++)
                       {
                           Node child2 = child1.getChildNodes().item(j);
                           if (child2.getNodeName().equalsIgnoreCase("Database"))
                           {
                                DBConfig db = new DBConfig();
                                if (NodeUtils.HasAttribute(child2,"Type"))
                                {
                                    db.setType(NodeUtils.GetAttributeValue(child2,"Type",""));
                                }
                                if (NodeUtils.HasAttribute(child2,"Id"))
                                {
                                    db.setDbIdentifier(NodeUtils.GetAttributeValue(child2,"Id",""));
                                }
                                if (NodeUtils.HasAttribute(child2,"Url"))
                                {
                                    db.setUrl(NodeUtils.GetAttributeValue(child2,"Url",""));
                                }
                                if (NodeUtils.HasAttribute(child2,"User"))
                                {
                                    db.setUser(NodeUtils.GetAttributeValue(child2,"User",""));
                                }
                                if (NodeUtils.HasAttribute(child2,"Pass"))
                                {
                                    db.setPass(NodeUtils.GetAttributeValue(child2,"Pass",""));
                                }
                                if (NodeUtils.HasAttribute(child2,"Driver"))
                                {
                                    db.setDriver(NodeUtils.GetAttributeValue(child2,"Driver",""));
                                }
                                if (NodeUtils.HasAttribute(child2,"MaxConnections"))
                                {
                                    db.setMaxConnections(new Integer(NodeUtils.GetAttributeValue(child2,"MaxConnections","")).intValue());
                                }
                               if (_log.isDebugEnabled()) {
                                   _log.debug("Created a new database configuration:");
                                   _log.debug(" * ID:               {}", db.getDbIdentifier());
                                   _log.debug(" * Name:             {}", db.getName());
                                   _log.debug(" * URL:              {}", db.getUrl());
                                   _log.debug(" * User:             {}", db.getUser());
                                   _log.debug(" * Driver:           {}", db.getDriver());
                                   _log.debug(" * Max connections:  {}", db.getMaxConnections());
                               }
                                DBPool.AddDBConfig(db);
                           }
                       }
                   }
               }else if (child1.getNodeName().equalsIgnoreCase("Package"))
                {
//					if (NodeUtils.HasAttribute(child1,"Name"))
//					{
//						this.setPackageName(NodeUtils.GetAttributeValue(child1,"Name",""));
//					}
                }else if (child1.getNodeName().equalsIgnoreCase("Models"))
                {
                    if (child1.hasChildNodes())
                    {
                        for (int j=0;j<child1.getChildNodes().getLength();j++)
                        {
                            Node child2 = child1.getChildNodes().item(j);
                            if (child2.getNodeName().equalsIgnoreCase("Data_Model"))
                            {
                                XFTDataModel model = new XFTDataModel();
                                if (NodeUtils.HasAttribute(child2,"File_Name"))
                                {
                                    model.setFileName(NodeUtils.GetAttributeValue(child2,"File_Name",""));
                                }
                                if (NodeUtils.HasAttribute(child2,"File_Location"))
                                {
                                    String file_location = NodeUtils.GetAttributeValue(child2,"File_Location","");
                                    if (!FileUtils.IsAbsolutePath(file_location))
                                    {
                                        file_location = sourceDir + file_location;
                                    }
                                    model.setFileLocation(file_location);
                                }
                                if (NodeUtils.HasAttribute(child2,"DB"))
                                {
                                    model.setDb(NodeUtils.GetAttributeValue(child2,"DB",""));
                                }
                                if (NodeUtils.HasAttribute(child2,"package"))
                                {
                                    model.setDb(NodeUtils.GetAttributeValue(child2,"package",""));
                                }
                                try {
                                    model.setSchema();
                                } catch (XFTInitException e) {
                                    e.printStackTrace();
                                } catch (ElementNotFoundException e) {
                                    e.printStackTrace();
                                }
                                if (_log.isDebugEnabled()) {
                                    _log.debug("Adding new data model from file {}: ", model.getFileName());
                                    _log.debug(" * Elements: {}", Joiner.on(", ").join(model.getElementNames()));
                                }
                                DATA_MODELS.put(model.getFileName(),model);
                            }
                        }
                    }
                }
            }
        }
    }

    @Override
    public String toString()
    {
        Document doc = toXML();
        return XMLUtils.DOMToString(doc);
    }

    public Document toXML()
    {
        XMLWriter writer = new XMLWriter();
        Document doc =writer.getDocument();
        Node main = doc.createElement("xdat-manager");
        Iterator schemas =  GetSchemas().iterator();
        while (schemas.hasNext())
        {
            XFTSchema schema = (XFTSchema)schemas.next();
            main.appendChild(schema.toXML(doc));
        }
        doc.appendChild(main);
        return doc;
    }

    private void manageAddins() throws ElementNotFoundException,XFTInitException,Exception
    {
        //XFT.LogCurrentTime("MANAGER ADD_INS:1","ERROR");

        ArrayList histories = new ArrayList();
        Iterator addIns = getAddInElements().iterator();
        while (addIns.hasNext())
        {
            XFTElement addIn = (XFTElement)addIns.next();
            if (_log.isDebugEnabled()) {
                _log.debug("Found add-in element: {}", addIn.getName());
            }
            if (addIn.getAddin().equalsIgnoreCase("global"))
            {
                //Every data element should have a reference to this element
                Iterator schemas =  GetSchemas().iterator();
                while (schemas.hasNext())
                {
                    XFTSchema schema = (XFTSchema)schemas.next();
                    Iterator elements = schema.getSortedElements().iterator();
                    while (elements.hasNext())
                    {
                        XFTElement element = (XFTElement)elements.next();
                        if (element.getAddin() == null || element.getAddin().equalsIgnoreCase(""))
                        {
                            XFTElement clone = addIn.clone(element,false);
                            clone.setExtensionType(null);
                            clone.setExtension(false);
                            clone.setAddin("meta");
                            clone.setSkipSQL(element.isSkipSQL());

                            XFTReferenceField field = XFTReferenceField.GetEmptyRef();
                            field.setName("meta");
                            field.setXMLType(clone.getType());
                            field.setFullName("meta");
                            field.setMinOccurs("0");
                            field.setExpose("false");
                            field.getRelation().setOnDelete("SET NULL");
                            field.setParent(element);
                            field.setSequence(1001);
                            String parentType= element.getType().getLocalType();
                            if (element.getSqlElement()!=null && element.getSqlElement().getName()!=null){
                                parentType=element.getSqlElement().getName();
                            }
                            field.getSqlField().setSqlName(parentType +"_info");
                            element.addField(field);


                            schema.addElement(clone);
                        }
                    }
                }
            }else if (addIn.getAddin().equalsIgnoreCase("local"))
            {
//				Every data element should have an additional table of this type
                Iterator schemas =  GetSchemas().iterator();
                while (schemas.hasNext())
                {
                    XFTSchema schema = (XFTSchema)schemas.next();
                    Iterator elements = schema.getSortedElements().iterator();
                    while (elements.hasNext())
                    {
                        XFTElement element = (XFTElement)elements.next();
                        if (element.getAddin() == null || element.getAddin().equalsIgnoreCase(""))
                        {
                            XFTElement clone = addIn.clone(element,false);
                            schema.addElement(clone);
                        }
                    }
                }
            }else if (addIn.getAddin().equalsIgnoreCase("history"))
            {
                histories.add(addIn);
            }else if (addIn.getAddin().equalsIgnoreCase("extension"))
            {

            }
        }

        _log.debug("Initializing the XFTReferenceManager");
        XFTReferenceManager.init();

        Iterator hist = histories.iterator();
        while (hist.hasNext())
        {
            XFTElement addIn = (XFTElement)hist.next();
//			Every data element should have an additional table of this type with its rows included
            Iterator schemas =  GetSchemas().iterator();
            while (schemas.hasNext())
            {
                XFTSchema schema = (XFTSchema)schemas.next();
                Iterator elements = schema.getSortedElements().iterator();
                while (elements.hasNext())
                {
                    XFTElement element = (XFTElement)elements.next();
                    GenericWrapperElement wrapE = (GenericWrapperElement)GenericWrapperFactory.GetInstance().wrapElement(element);

                    if (element.getAddin() == null || element.getAddin().equalsIgnoreCase(""))
                    {
                        XFTElement clone = addIn.clone(element,true);
                        clone.setExtension(false);

                        clone.setSkipSQL(element.isSkipSQL());

                        final List<GenericWrapperField> fields=(List<GenericWrapperField>) wrapE.getAllFieldsWAddIns(false,false);
                        for(GenericWrapperField field:fields)
                        {
                            if (field.isReference())
                            {
                                XFTReferenceI ref = field.getXFTReference();
                                try {
                                    if (! ref.isManyToMany())
                                    {
                                        Iterator specs = ((XFTSuperiorReference)ref).getKeyRelations().iterator();
                                        while (specs.hasNext())
                                        {
                                            XFTRelationSpecification spec = (XFTRelationSpecification)specs.next();
                                            XFTField cloneField = XFTDataField.GetEmptyField();
                                            cloneField.setMinOccurs("0");
                                            cloneField.setRequired("");
                                            cloneField.setParent(clone);
                                            cloneField.setUnique("false");
                                            cloneField.setUniqueComposite("false");
                                            cloneField.setName(spec.getLocalCol());
                                            cloneField.setFullName(spec.getLocalCol());
                                            cloneField.setXMLType(spec.getSchemaType());
                                            clone.addField(cloneField);
                                        }
                                    }else{
                                    	System.out.println();
                                    }
                                } catch (RuntimeException e) {
                                    throw new RuntimeException("Error managing XDAT add-ins for element(" + wrapE.getFullXMLName() + ") field(" + field.getXMLPathString("") + ")");
                                }
                            }else
                            {
                                if (GenericWrapperField.IsLeafNode(field.getWrapped()))
                                {
                                    XFTField cloneField = XFTDataField.GetEmptyField();
                                    cloneField.setMinOccurs("0");
                                    cloneField.setRequired("");
                                    cloneField.setParent(clone);
                                    cloneField.setUnique("false");
                                    cloneField.setUniqueComposite("false");
                                    cloneField.setSize(field.getSize());
                                    cloneField.setName(field.getSQLName());
                                    cloneField.setFullName(field.getSQLName());
                                    cloneField.getSqlField().setSqlName(field.getSQLName());
                                    cloneField.setXMLType(field.getXMLType());
                                    clone.addField(cloneField);
                                }
                            }
                        }


                        XFTField cloneField = XFTDataField.GetEmptyField();
                        cloneField.setMinOccurs("0");
                        cloneField.setRequired("");
                        cloneField.setParent(clone);
                        cloneField.setUnique("false");
                        cloneField.setUniqueComposite("false");
                        cloneField.setSize("");
                        cloneField.setName("change_id");
                        cloneField.setFullName("xft_version");
                        cloneField.getSqlField().setSqlName("xft_version");
                        cloneField.setXMLType(new XMLType("xs:integer",schema));
                        clone.addField(cloneField);

                        schema.addElement(clone);
                    }
                }
            }
        }
    }

    private ArrayList getAddInElements()
    {
        ArrayList al = new ArrayList();
        Iterator schemas =  GetSchemas().iterator();
        while (schemas.hasNext())
        {
            XFTSchema schema = (XFTSchema)schemas.next();
            Iterator elements = schema.getSortedElements().iterator();
            while (elements.hasNext())
            {
                XFTElement element = (XFTElement)elements.next();
                if (_log.isDebugEnabled()) {
                    _log.debug("Checking XFT element {} to see if it's an add-in element", element.getName());
                }
                if (element.getAddin() != null && !element.getAddin().equalsIgnoreCase(""))
                {
                    if (_log.isDebugEnabled()) {
                        _log.debug("XFT element {} is an add-in element, adding it to the list", element.getName());
                    }
                    al.add(element);
                }
            }
        }
        if (_log.isDebugEnabled()) {
            _log.debug("Found a total of {} add-in elements", al.size());
        }
        return al;
    }

    /**
     * ArrayList of GenericWrapperElements
     * @return
     * @throws XFTInitException
     * @throws ElementNotFoundException
     * @throws Exception
     */
    public ArrayList getOrderedElements() throws XFTInitException,ElementNotFoundException,Exception
    {
        ArrayList al = new ArrayList();
        Iterator schemas =  GetSchemas().iterator();
        XFTElementSorter sorter = new XFTElementSorter();
        while (schemas.hasNext())
        {
            XFTSchema schema = (XFTSchema)schemas.next();
            Iterator elements = schema.getSortedElements().iterator();
            while (elements.hasNext())
            {
                XFTElement element = (XFTElement)elements.next();
                sorter.addElement((GenericWrapperElement)GenericWrapperFactory.GetInstance().wrapElement(element));
            }
        }

        al.addAll(sorter.getElements());

        al.trimToSize();
        return al;
    }

    /**
     * ArrayList of GenericWrapperElements
     * @return
     * @throws XFTInitException
     * @throws ElementNotFoundException
     * @throws Exception
     */
    public ArrayList getAllElements() throws XFTInitException
    {
        ArrayList al = new ArrayList();
        Iterator schemas =  GetSchemas().iterator();
        while (schemas.hasNext())
        {
            XFTSchema schema = (XFTSchema)schemas.next();
            Iterator elements = schema.getSortedElements().iterator();
            while (elements.hasNext())
            {
                XFTElement element = (XFTElement)elements.next();
                al.add(GenericWrapperFactory.GetInstance().wrapElement(element));
            }
        }

        al.trimToSize();
        return al;
    }
}

